import okcupyd


u = okcupyd.User()
p = u.quickmatch()


print('Questions for {0}'.format(p.username))
print('----------')
for q in p.questions:
    if 'marijuana' in q.text.lower():
        print(q.text)
        print(q.user_answer)
        if q.explanation and len(q.explanation):
            print(q.explanation)
        print('')



import sys

import mock

import okcupyd


@mock.patch.object(sys, 'argv', ['okcupyd'])
@mock.patch.object(sys, 'exit')
@mock.patch.object(sys, 'exit')
@mock.patch.object(okcupyd, 'start_ipython')
def test_interactive(*args):
    okcupyd.parse_args_and_run()
    with mock.patch.object(sys, 'argv', ['okcupyd', '-v']):
        okcupyd.parse_args_and_run()



import mock
import pytest

from okcupyd.search import SearchManager, SearchParameterBuilder, search
from okcupyd.profile import Profile
from . import util


@pytest.fixture
@util.use_cassette(cassette_name='search_startup')
def search_manager():
    return SearchManager(looking_for='everybody')


@util.use_cassette(cassette_name='search_age_filter')
def test_age_filter(search_manager):
    age = 22
    search_manager.set_options(age_min=age, age_max=age)
    profile, = search_manager.get_profiles(count=1)
    assert profile.age == age


@util.use_cassette(cassette_name='search_count')
def test_count_variable(search_manager, request):
    search_manager.set_options(looking_for='everybody')
    profiles = search_manager.get_profiles(count=14)
    assert len(profiles) == 14

    for profile in profiles:
        profile.username
        profile.age
        profile.location
        profile.match_percentage
        profile.enemy_percentage
        profile.id
        profile.rating
        profile.contacted


@util.use_cassette(cassette_name='search_location_filter')
def test_location_filter(search_manager):
    location = 'Portland, OR'
    search_manager.set_options(location=location, radius=25)
    profile, = search_manager.get_profiles(count=1)
    assert profile.location == 'Portland, OR'


@util.use_cassette(cassette_name='search_function')
def test_search_function():
    profile, = search(count=1)
    assert isinstance(profile, Profile)

    profile.username
    profile.age
    profile.location
    profile.match_percentage
    profile.enemy_percentage
    profile.id
    profile.rating
    profile.contacted


@util.use_cassette
def test_search_manager_iter():
    search_manager = SearchManager(looking_for='everybody',
                                   religion='buddhist', age_min=25, age_max=25,
                                   location='new york, ny', keywords='bicycle')
    for profile in search_manager:
        assert isinstance(profile, Profile)


@mock.patch('okcupyd.helpers.get_locid', return_value=2)
def test_construction_of_all_search_parameters(mock_get_locid):
    spb = SearchParameterBuilder()
    spb.set_options(location='new york, ny', religion='buddhist',
                    height_min=66, height_max=68, looking_for='everybody',
                    smokes=['no', 'trying to quit'], age_min=18, age_max=24,
                    radius=12, order_by='MATCH', last_online=1234125,
                    status='single', drugs=['very_often', 'sometimes'],
                    job=['retired'], education=['high school'],
                    income='less than $20,000', monogomy='monogamous',
                    diet='vegan', ethnicity=['asian', 'middle eastern'],
                    pets=['owns dogs', 'likes cats'], kids=['has a kid'])
    spb.build(mock.Mock(), count=10)


def test_empty_string_looking_for():
    SearchParameterBuilder(looking_for='').build(mock.Mock())



import time
import logging

import requests

from .errors import AuthenticationError
from . import settings
from . import util
from . import profile


log = logging.getLogger(__name__)


class Session(requests.Session):

    default_login_headers = {
        'user-agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_4) '
                       'AppleWebKit/537.36 (KHTML, like Gecko) '
                       'Chrome/37.0.2062.94 '
                       'Safari/537.36')
    }

    @classmethod
    def login(cls, username=None, password=None, headers=None):
        # settings.USERNAME and settings.PASSWORD should not be made
        # the defaults to their respective arguments because doing so
        # would prevent this function from picking up any changes made
        # to those values after import time.
        username = username or settings.USERNAME
        password = password or settings.PASSWORD
        session = cls()
        credentials = {
            'username': username,
            'password': password,
            'okc_api': 1
        }
        login_response = session.post('https://www.okcupid.com/login',
                                      data=credentials,
                                      headers=headers or cls.default_login_headers)
        log_in_name = login_response.json()['screenname']
        if log_in_name is None:
            raise AuthenticationError('Could not log in as {0}'.format(username))
        if log_in_name.lower() != username.lower():
            log.warning('Expected to log in as {0} but got {1}'.format(username,
                                                                       log_in_name))
        log.debug(login_response.content.decode('utf8'))
        session.log_in_name = log_in_name
        session.headers.update(cls.default_login_headers)
        return session

    def __init__(self, *args, **kwargs):
        super(Session, self).__init__(*args, **kwargs)
        self.timestamp = -settings.DELAY

    def _throttle(self):
        while time.clock() - self.timestamp < settings.DELAY:
            time.sleep(.5)
        self.timestamp = time.clock()

    def post(self, *args, **kwargs):
        self._throttle()
        response = super(Session, self).post(*args, **kwargs)
        response.raise_for_status()
        return response

    def get(self, *args, **kwargs):
        self._throttle()
        response = super(Session, self).get(*args, **kwargs)
        response.raise_for_status()
        return response

    def okc_get(self, path, *args, **kwargs):
        return self.get('{0}{1}'.format(util.BASE_URI, path), *args, **kwargs)

    def okc_post(self, path, *args, **kwargs):
        return self.post('{0}{1}'.format(util.BASE_URI, path), *args, **kwargs)

    def get_profile(self, username):
        return profile.Profile(self, username)

    def get_current_user_profile(self):
        return self.get_profile(self.log_in_name)



import datetime

from . import util
from okcupyd import profile
from okcupyd import session
from okcupyd import User
from okcupyd.util import cached_property


@util.use_cassette(cassette_name='test_profile_essays')
def test_profile_essays():
    user_profile = profile.Profile(session.Session.login(), 'FriedLiverAttack')
    assert user_profile.essays.self_summary


@util.use_cassette(cassette_name='test_rate_profile')
def test_rate_profile():
    profile = User().quickmatch()
    profile.rate(5)
    profile.refresh()
    assert profile.rating == 5


@util.use_cassette
def test_rate_profile_1_stars():
    profile = User().quickmatch()
    profile.rate(1)
    assert User().get_profile(profile.username).rating == profile.rating
    assert profile.rating == 1


@util.use_cassette
def test_profile_properties():
    profile = User().quickmatch()
    assert 0 <= profile.match_percentage <= 100
    assert 0 <= profile.enemy_percentage <= 100
    assert profile.responds
    assert profile.contacted == False # We want it to be false, not falsy


@util.use_cassette(cassette_name='test_profile_contacted')
def test_contacted():
    profile = User().quickmatch()
    profile.message('test')
    assert bool(profile.contacted)
    assert isinstance(profile.contacted, datetime.datetime)


@util.use_cassette(cassette_name='test_profile_on_inbox_correspondent')
def test_contacted_on_inbox_correspondent():
    profile = User().outbox[-1].correspondent_profile
    assert bool(profile.contacted)
    assert isinstance(profile.contacted, datetime.datetime)
    for prop_name, _ in cached_property.get_cached_properties(profile):
        getattr(profile, prop_name)



from okcupyd.db import model, with_txn


def add_message_to_campaign_and_template_by_okc_id_no_txn(session, okc_id,
                                                          campaign, template):
    message = model.Message.find_no_txn(session, okc_id, id_key='id')
    message.template = template
    message.campaign = campaign
    return message


add_message_to_campaign_and_template_by_okc_id = with_txn(
    add_message_to_campaign_and_template_by_okc_id_no_txn
)


from sqlalchemy import Column
from sqlalchemy import String

from okcupyd.db import OKCBase


class User(OKCBase):

    __tablename__ = "user"

    handle = Column(String, nullable=False)
    age = Column(String, nullable=False)
    location = Column(String, nullable=False)



import logging
import itertools
import os

import mock
import contextlib2 as contextlib
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from vcr.cassette import Cassette, CassetteContextDecorator

from . import util
from okcupyd import db, settings
from okcupyd import util as okcupyd_util
from okcupyd.db import model, adapters


log = logging.getLogger(__name__)
BREAK_ON_EXCEPTION = False


def pytest_addoption(parser):
    okcupyd_util.add_command_line_options(parser.addoption, use_short_options=False)
    parser.addoption('--live', dest='skip_vcrpy', action='store_true', default=False,
                     help="Skip the patching of http requests in tests. "
                     "USE WITH CAUTION. This will interact with the okcupid "
                     "website and send messages with any provided user credentials.")
    parser.addoption('--scrub', dest='scrub', action='store_true', default=False,
                     help="USE WITH CAUTION. Don't scrub PII from "
                     "http requests/responses. This is useful for recording cassetes.")
    parser.addoption('--resave', dest='resave',
                     action='store_true', default=False,
                     help="Resave cassettes. Use to retoractively scrub cassettes.")
    parser.addoption('--cassette-mode', dest='cassette_mode', action='store',
                     default='once', help="Accept new requests in all tests.")
    parser.addoption('--record', dest='record', action='store_true',
                     default=False, help="Re-record cassettes for all tests.")
    parser.addoption('--break-exc', dest='break_on_exception', action='store_true',
                     default=False, help="Enter ipdb if an exception is "
                     "encounteredin a test.")


def patch(option, *patches, **kwargs):
    negate = kwargs.get('negate', False)
    @pytest.yield_fixture(autouse=True, scope='session')
    def patch_conditionally(request):
        condition = bool(request.config.getoption(option))
        if negate: condition = not condition
        if condition:
            with contextlib.ExitStack() as exit_stack:
                for patch in patches:
                    exit_stack.enter_context(patch)
                yield
        else:
            yield
    return patch_conditionally


patch_settings = patch('credentials_modules',
                       mock.patch.object(settings, 'USERNAME', 'username'),
                       mock.patch.object(settings, 'PASSWORD', 'password'),
                       mock.patch.object(settings, 'AF_USERNAME', 'username'),
                       mock.patch.object(settings, 'AF_PASSWORD', 'password'),
                       mock.patch.object(settings, 'DELAY', 0), negate=True)
patch_save = patch('resave',
                   mock.patch.object(
                       Cassette, '_save',
                       okcupyd_util.n_partialable(Cassette._save)(force=True)))
patch_use_cassette_enabled = patch('skip_vcrpy',
                                   mock.patch.object(CassetteContextDecorator,
                                                     '__enter__'),
                                   mock.patch.object(CassetteContextDecorator,
                                                     '__exit__'))
patch_vcrpy_filters = patch('scrub',
                            mock.patch.object(util, 'SHOULD_SCRUB', False),
                            negate=True)

patch_break_on_exception = patch('break_on_exception',
                                 mock.patch('tests.conftest.BREAK_ON_EXCEPTION',
                                            True))

original_cassette_enter = CassetteContextDecorator.__enter__
def new_cassette_enter(self):
    path, _ = self._args_getter()
    try:
        os.remove(path)
    except:
        pass
    original_cassette_enter(self)
patch_record = patch('record', mock.patch.object(CassetteContextDecorator,
                                                 '__enter__',
                                                 new_cassette_enter))

@pytest.fixture(autouse=True, scope='session')
def set_vcr_options(request):
    util.okcupyd_vcr.record_mode = request.config.getoption('cassette_mode')


@pytest.fixture(autouse=True, scope='session')
def process_command_line_args(request):
    okcupyd_util.handle_command_line_options(request.config.option)


@pytest.fixture(scope='function')
def engine(request):
    return create_engine('sqlite://', echo=request.config.getoption('echo'))


@pytest.fixture(scope='function')
def session(engine):
    new_kwargs = db.Session.kw.copy()
    new_kwargs['bind'] = engine
    return sessionmaker(**new_kwargs)


@pytest.yield_fixture(autouse=True, scope='function')
def setup_db(engine, session):
    with mock.patch.object(db, 'Session', session):
        orig_engine = db.Base.metadata.bind
        db.Base.metadata.bind = engine
        db.Base.metadata.drop_all()
        db.Base.metadata.create_all()
        yield
        db.Base.metadata.bind = orig_engine


@pytest.fixture
def T(mock_profile_builder, mock_message_thread_builder, mock_message_builder):
    class testing(object): pass
    testing.ensure = mock.Mock()
    testing.build_mock = mock.Mock()
    testing.factory = mock.Mock()
    def ensure_thread_model_resembles_okcupyd_thread(thread_model,
                                                     okcupyd_thread):
        assert thread_model.okc_id == okcupyd_thread.id
        ensure_user_model_resembles_okcupyd_profile(
            thread_model.initiator, okcupyd_thread.initiator
        )
        ensure_user_model_resembles_okcupyd_profile(
            thread_model.respondent, okcupyd_thread.respondent
        )
        for pair in zip(thread_model.messages, okcupyd_thread.messages):
            ensure_message_model_resembles_okcupyd_message(*pair)
            assert len(thread_model.messages) == len(okcupyd_thread.messages)


    def ensure_message_model_resembles_okcupyd_message(message_model,
                                                       okcupyd_message):
        assert message_model.okc_id == okcupyd_message.id
        assert message_model.sender.handle == okcupyd_message.sender
        assert message_model.recipient.handle == okcupyd_message.recipient
        assert message_model.text == okcupyd_message.content



    def ensure_user_model_resembles_okcupyd_profile(user_model,
                                                    okcupyd_profile):
        assert user_model.handle == okcupyd_profile.username

    testing.ensure.user_model_resembles_okcupyd_profile = \
        ensure_user_model_resembles_okcupyd_profile
    testing.ensure.message_model_resembles_okcupyd_message = \
        ensure_message_model_resembles_okcupyd_message
    testing.ensure.thread_model_resembles_okcupyd_thread = \
        ensure_thread_model_resembles_okcupyd_thread

    testing.build_mock.thread = mock_message_thread_builder
    testing.build_mock.message = mock_message_builder
    testing.build_mock.profile = mock_profile_builder

    def build_message_thread():
        message_thread = testing.build_mock.thread()
        return adapters.ThreadAdapter(message_thread).get_thread()

    def build_user(username):
        profile = testing.build_mock.profile(username)
        return adapters.UserAdapter(profile).get()

    testing.factory.message_thread = build_message_thread
    testing.factory.user = build_user
    return testing


@pytest.fixture
def mock_profile_builder():
    counter = itertools.count()
    next(counter)
    username_to_profile = {}
    def _build_mock_profile(username='username', age=30, id=None,
                            location='San Francisco, CA', **kwargs):
        if username in username_to_profile:
            return username_to_profile[username]
        if id is None: id = next(counter)
        mock_profile = mock.MagicMock(id=id or next(counter), location=location,
                                      age=age, username=username, **kwargs)
        username_to_profile[username] = mock_profile
        return mock_profile
    return _build_mock_profile


@pytest.fixture
def mock_message_builder():
    counter = itertools.count()
    next(counter)
    def _build_mock_message(id=None, sender='sender', recipient='recipient',
                            content='content'):
        if id == None: id = next(counter)
        assert isinstance(sender, str)
        assert isinstance(recipient, str)
        return mock.MagicMock(id=id, sender=sender, recipient=recipient,
                              content=content)
    return _build_mock_message


@pytest.fixture
def mock_message_thread_builder(mock_message_builder, mock_profile_builder):
    counter = itertools.count()
    next(counter)
    def _build_mock_message_thread(id=None, message_count=2,
                                   initiator='initiator',
                                   respondent='respondent', **kwargs):
        if id == None: id = next(counter)
        messages = [mock_message_builder(content='{0}'.format(i),
                                         sender=respondent
                                         if i % 2 else initiator,
                                         recipient=initiator
                                         if i % 2 else respondent)
                    for i in range(message_count)]
        message_thread = mock.MagicMock(id=id, messages=messages, **kwargs)
        message_thread.initiator = mock_profile_builder(initiator)
        message_thread.respondent = mock_profile_builder(respondent)
        return message_thread
    return _build_mock_message_thread


def pytest_exception_interact():
    models = []
    for key in dir(model):
        prospect = getattr(model, key)
        if isinstance(prospect, type) and issubclass(prospect, db.Base):
            models.append(prospect)
    model_class_to_instances = {}
    for model_class in models:
        model_class_to_instances[model_class] = {
            instance.okc_id: instance for instance in model_class.query()
        }
    log.info(model_class_to_instances)
    if BREAK_ON_EXCEPTION:
        import ipdb; ipdb.set_trace()


from . import util
from okcupyd import User
from okcupyd.profile import Profile


@util.use_cassette
def test_unicode_in_messages_threads():
    new_user_instance = User()
    assert len(new_user_instance.inbox.items) >= 30
    assert len(new_user_instance.outbox.items) >= 30


@util.use_cassette
def test_inbox_refresh():
    User().inbox.refresh()


@util.use_cassette
def test_initiator_and_respondent():
    assert isinstance(User().inbox[0].initiator, Profile)
    assert isinstance(User().inbox[0].respondent, Profile)


import datetime
import itertools
import logging

from lxml import html
import simplejson

from . import helpers
from . import util
from .xpath import xpb


log = logging.getLogger(__name__)


class LookingFor(object):

    _looking_for_xpb = xpb.div.with_classes('text', 'what_i_want')

    def __init__(self, profile_tree_owner):
        self._profile_tree_owner

    @util.cached_property
    def raw_fields(self):
        li_elements = self._looking_for_xpb.li.apply_(
            self._profile_tree_owner._profile_tree
        )
        return {li.attrib['id'].split('_')[1]: li.text_content()
                for li in li_elements}

    @util.cached_property
    def ages(self):
        return self.raw_fields.get('ages')

    @util.cached_property
    def single(self):
        return bool(self.raw_fields.get('ages', None))

    @util.cached_property
    def near_me(self):
        return 'near' in self.raw_field.get('near', '').lower()

    @util.cached_property
    def looking_for(self):
        return 'near' in self.raw_field.get('near', '').lower()


class Essays(object):

    @staticmethod
    def build_essay_property(essay_index, essay_name):
        essay_xpb = xpb.div(id='essay_{0}'.format(essay_index))
        essay_text_xpb = essay_xpb.div.with_class('text').div.with_class('essay')
        @property
        def essay(self):
            try:
                essay_text = essay_text_xpb.get_text_(self._essays).strip()
            except IndexError:
                return None
            if essay_name not in self._short_name_to_title:
                self._short_name_to_title[essay_name] = helpers.replace_chars(
                    essay_xpb.a.with_class('essay_title').get_text_(
                        self._profile_tree
                    )
                )
            return essay_text

        @essay.setter
        def set_essay_text(self, essay_text):
            self._submit_essay(essay_index, essay_text)

        return set_essay_text

    @classmethod
    def _init_essay_properties(cls):
        for essay_index, essay_name in enumerate(cls.essay_names):
            setattr(cls, essay_name,
                    cls.build_essay_property(essay_index, essay_name))

    _essays_xpb = xpb.div(id='main_column')
    essay_names = ['self_summary', 'my_life', 'good_at', 'people_first_notice',
                    'favorites', 'six_things', 'think_about', 'friday_night',
                    'private_admission', 'message_me_if']

    def __init__(self, session, profile_tree, profile_tree_getter=None):
        self._session = session
        self._profile_tree = profile_tree
        self._profile_tree_getter = profile_tree_getter
        self._short_name_to_title = {}

    @property
    def short_name_to_title(self):
        for i in self: pass # Make sure that all essays names have been retrieved
        return self._short_name_to_title


    @util.cached_property
    def _essays(self):
        return self._essays_xpb.one_(self._profile_tree)

    @util.cached_property
    def _authcode(self):
        return helpers.get_authcode(self._profile_tree)

    def _submit_essay(self, essay_id, essay_body):
        self._session.post('https://www.okcupid.com/profileedit2', data={
            "essay_id": essay_id,
            "essay_body": essay_body,
            "authcode": self._authcode,
            "okc_api": 1
        })
        self.refresh()

    def refresh(self, profile_tree=None):
        if profile_tree is None:
            assert self._profile_tree_getter, (
                "Must provide a profile_tree or a profile_tree_getter"
            )
            self._profile_tree = self._profile_tree_getter()
        util.cached_property.bust_caches(self)

    def __iter__(self):
        for essay_name in self.essay_names:
            yield getattr(self, essay_name)


Essays._init_essay_properties()


class Question(object):

    def __init__(self, id_, text, user_answer, explanation):
        self.id = id_
        self.text = text
        self.user_answer = user_answer
        self.explanation = explanation

    def __repr__(self):
        return '<Question: {0}>'.format(self.text)


class QuestionFetcher(object):

    step = 10

    @classmethod
    def build(cls, session, username):
        return cls(QuestionHTMLFetcher(session, username),
                   QuestionProcessor(session))

    _page_data_xpb = xpb.div.with_class('pages_data')
    _current_page_xpb = _page_data_xpb.input(id='questions_pages_page').\
                        select_attribute_('value')
    _total_page_xpb = _page_data_xpb.input(id='questions_pages_total').\
                      select_attribute_('value')

    def __init__(self, html_fetcher, processor):
        self._html_fetcher = html_fetcher
        self._processor = processor

    def _pages_left(self, tree):
        return (int(self._current_page_xpb.one_(tree)) <
                int(self._total_page_xpb.one_(tree)))

    def fetch(self, start_at=0, get_at_least=None, id_to_existing=None):
        current_offset = start_at + 1
        generators = []
        while True:
            tree = html.fromstring(self._html_fetcher.fetch(current_offset))
            generators.append(self._processor.process(tree))
            if not self._pages_left(tree):
                break
            current_offset = current_offset + self.step
        return itertools.chain(*generators)


class QuestionHTMLFetcher(object):

    def __init__(self, session, username, step=10, **additional_parameters):
        self._session = session
        self._username = username
        self.step = step
        self._additional_parameters = additional_parameters

    @property
    def _uri(self):
        return 'profile/{0}/questions'.format(self._username)

    def _query_params(self, start_at):
        parameters = {'low': start_at, 'leanmode': 1}
        parameters.update(self._additional_parameters)
        return parameters

    def fetch(self, start_at):
        response = self._session.okc_get(self._uri,
                                         params=self._query_params(start_at))
        return response.content.decode('utf8')


class QuestionProcessor(object):

    _question_xpb = xpb.div.with_class('question')

    def __init__(self, session, id_to_existing=None):
        self._session = session
        self.id_to_existing = id_to_existing or {}

    def _process_question_element(self, question_element):
        id_ = question_element.attrib['data-qid']
        text = xpb.div.with_class('qtext').p.get_text_(question_element).strip()
        answer = None
        explanation = None
        if 'not_answered' not in question_element.attrib['class']:
            answer = xpb.span.attribute_contains('id', 'answer_target').\
                     get_text_(question_element).strip()

            explanation = xpb.div.span.with_class('note').\
                          xget_text_(question_element).strip()
        return Question(id_, text, answer, explanation)

    def process(self, tree):
        for question_element in self._question_xpb.apply_(tree):
                yield self._process_question_element(question_element)


class Profile(object):

    def __init__(self, session, username, **kwargs):
        self._session = session
        self.username = username
        self._question_fetcher = QuestionFetcher.build(session, username)
        self.questions = util.Fetchable(self._question_fetcher)

    def refresh(self, reload=True):
        util.cached_property.bust_caches(self)
        if reload:
            return self._profile_tree

    @property
    def is_logged_in_user(self):
        return self._session.log_in_name.lower() == self.username.lower()

    @util.cached_property
    def _profile_response(self):
        return self._session.get(
            'https://www.okcupid.com/profile/{0}'.format(self.username)
        ).content

    @util.cached_property
    def _profile_tree(self):
        return html.fromstring(self._profile_response)

    def message_request_parameters(self, content, thread_id):
        return {
            'ajax': 1,
            'sendmsg': 1,
            'r1': self.username,
            'body': content,
            'threadid': thread_id,
            'authcode': self.authcode,
            'reply': 1 if thread_id else 0,
            'from_profile': 1
        }

    @util.cached_property
    def authcode(self):
        return helpers.get_authcode(self._profile_tree)

    @util.cached_property
    def picture_uris(self):
        pics_request = self._session.okc_get(
            'profile/{0}/photos?cf=profile'.format(self.username)
        )
        pics_tree = html.fromstring(pics_request.content.decode('utf8'))
        return xpb.div(id='album_0').img.select_attribute_('src', pics_tree)

    _rating_xpb = xpb.div(id='rating').ul(id='personality-rating').li.\
                  with_class('current-rating')

    @util.cached_property
    def rating(self):
        if self.is_logged_in_user: return 0
        rating_style = self._rating_xpb.select_attribute_('style').one_(
            self._profile_tree
        )
        width_percentage = int(''.join(c for c in rating_style if c.isdigit()))
        return width_percentage // 20

    _contacted_xpb = xpb.div(id='actions').div.with_classes('tooltip_text',
                                                               'hidden')

    @util.cached_property
    def contacted(self):
        try:
            contacted_span = self._contacted_xpb.span.with_class('fancydate').\
                             one_(self._profile_tree)
        except:
            return False
        else:
            timestamp = contacted_span.attrib['id'].split('_')[-1][:-2]
            return datetime.datetime.fromtimestamp(int(timestamp[:10]))

    @util.cached_property
    def responds(self):
        contacted_text = self._contacted_xpb.get_text_(self._profile_tree).lower()
        if 'contacted' not in contacted_text:
            return contacted_text.strip().replace('replies ', '')

    @util.cached_property
    def id(self):
        if self.is_logged_in_user: return self._current_user_id
        return int(self._rating_xpb.select_attribute_('id').
                   one_(self._profile_tree).split('-')[-2])

    @util.cached_property
    def _current_user_id(self):
        return int(helpers.get_id(self._profile_tree))

    @util.cached_property
    def essays(self):
        return Essays(self._session, self._profile_tree,
                      profile_tree_getter=self.refresh)

    @util.cached_property
    def age(self):
        return int(xpb.span(id='ajax_age').get_text_(self._profile_tree).strip())

    _percentages_and_ratings_xpb = xpb.div(id='percentages_and_ratings')

    @util.cached_property
    def match_percentage(self):
        return int(self._percentages_and_ratings_xpb.
                   div.with_class('match').
                   span.with_class('percent').
                   get_text_(self._profile_tree).strip('%'))

    @util.cached_property
    def enemy_percentage(self):
        return int(self._percentages_and_ratings_xpb.
                   div.with_class('enemy').
                   span.with_class('percent').
                   get_text_(self._profile_tree).strip('%'))

    @util.cached_property
    def location(self):
        return xpb.span(id='ajax_location').get_text_(self._profile_tree)

    @util.cached_property
    def gender(self):
        return xpb.span(id='ajax_gender').get_text_(self._profile_tree)

    @util.cached_property
    def orientation(self):
        return xpb.dd(id='ajax_orientation').get_text_(self._profile_tree).strip()

    _details_xpb = xpb.div(id='profile_details')

    @util.cached_property
    def details(self):
        details = {}
        details_div = self._details_xpb.one_(self._profile_tree)
        for dl in details_div.iter('dl'):
            title = dl.find('dt').text
            item = dl.find('dd')
            if title == 'Last Online' and item.find('span') is not None:
                details[title.lower()] = item.find('span').text.strip()
            elif title.lower() in details and len(item.text):
                details[title.lower()] = item.text.strip()
            else:
                continue
            details[title.lower()] = helpers.replace_chars(details[title.lower()])
        return details

    @util.n_partialable
    def message(self, message, thread_id=None):
        return_value =  helpers.Messager(
            self._session
        ).send(self.username, message,
                       self.authcode, thread_id)
        self.refresh(reload=False)
        return return_value

    def rate(self, rating):
        parameters = {
            'voterid': self._current_user_id,
            'target_userid': self.id,
            'type': 'vote',
            'cf': 'profile2',
            'target_objectid': 0,
            'vote_type': 'personality',
            'score': rating,
        }
        headers = {
            'accept': 'text/javascript, text/html, application/xml, '
            'text/xml, */*',
            'accept-encoding': 'gzip,deflate',
            'accept-language': 'en-US,en;q=0.8',
            'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
            'origin': 'https://www.okcupid.com',
            'x-prototype-version': 1.7,
            'x-requested-with': 'XMLHttpRequest',
        }
        response = self._session.okc_post('vote_handler', data=parameters,
                                          headers=headers)
        response_json = response.json()
        log_function = log.info if response_json.get('status', False) \
                       else log.error
        log_function(simplejson.dumps({'rate_response': response_json,
                                       'sent_parameters': parameters,
                                       'headers': dict(self._session.headers)}))
        self.refresh(reload=False)

    def __repr__(self):
        return 'Profile("{0}")'.format(self.username)




import mock
import pytest

from .. import util
from okcupyd import User
from okcupyd.db import mailbox_sync, model, txn


@pytest.fixture
def mock_user():
    mock_user = mock.MagicMock()
    return mock_user


@pytest.fixture
def inbox_sync(mock_user):
    return mailbox_sync.MailboxSyncer(mock_user.inbox)


def set_inbox(inbox, value):
    inbox.__iter__ = value.__iter__
    inbox.items = value
    inbox.refresh.return_value = value


def test_mailbox_sync_creates_message_rows(T, inbox_sync, mock_user):
    initiator_one = 'first_initiator'
    respondent_one = 'respondent_one'
    respondent_two = 'respondent_two'
    initiator_two = 'other'
    set_inbox(mock_user.inbox, [
        T.build_mock.thread(initiator=initiator_one, respondent=respondent_one),
        T.build_mock.thread(initiator=initiator_one, respondent=respondent_two),
        T.build_mock.thread(initiator=initiator_two, respondent=respondent_two,
                            message_count=1)
    ])

    id_to_mock_thread = {t.id: t for t in mock_user.inbox.items}

    inbox_sync.sync()
    mock_user.inbox.refresh.assert_called_once_with()

    with txn() as session:
        message_threads = session.query(model.MessageThread).all()
        assert len(message_threads) == len(id_to_mock_thread)
        for message_thread in message_threads:
            T.ensure.thread_model_resembles_okcupyd_thread(
                message_thread,
                id_to_mock_thread[message_thread.id]
            )

    # Sync again and make sure that nothing has been updated.
    inbox_sync.sync()
    with txn() as session:
        message_threads = session.query(model.MessageThread).all()
        assert len(message_threads) == len(id_to_mock_thread)
        for message_thread in message_threads:
            T.ensure.thread_model_resembles_okcupyd_thread(
                message_thread,
                id_to_mock_thread[message_thread.okc_id]
            )

    # Add messages to existing threads
    mock_user.inbox.items[0].messages.append(T.build_mock.message(
        sender=respondent_one, recipient=initiator_one, content='final'
    ))

    mock_user.inbox.items[2].messages.append(T.build_mock.message(
        sender=initiator_two, recipient=respondent_two, content='last'
    ))

    # Sync and make sure that only the new messages appear.
    inbox_sync.sync()
    with txn() as session:
        message_threads = session.query(model.MessageThread).all()
        assert len(message_threads) == len(id_to_mock_thread)
        for message_thread in message_threads:
            T.ensure.thread_model_resembles_okcupyd_thread(
                message_thread,
                id_to_mock_thread[message_thread.okc_id]
            )

    # Add a new message thread and make sure that only it appears.
    set_inbox(mock_user.inbox, [
        T.build_mock.thread(initiator=initiator_one,
                            respondent=initiator_two)
    ] + mock_user.inbox.items)

    id_to_mock_thread = {t.id: t for t in mock_user.inbox.items}
    inbox_sync.sync()

    with txn() as session:
        message_threads = session.query(model.MessageThread).all()
        for message_thread in message_threads:
            T.ensure.thread_model_resembles_okcupyd_thread(
                message_thread,
                id_to_mock_thread[message_thread.okc_id]
            )

@util.use_cassette
def test_mailbox_sync_integration():
    user = User()
    user.quickmatch().message('test... sorry.')

    mailbox_sync.MailboxSyncer(user.outbox).sync()

    user_model = model.User.find(user.profile.id, id_key='okc_id')
    messages = model.Message.query(
        model.Message.sender_id == user_model.id
    )

    mailbox_sync.MailboxSyncer(user.outbox).sync()

    assert len(messages) == len(model.Message.query(
        model.Message.sender_id == user_model.id
    ))


"""
Search OKCupid profiles, return a list of matching profiles.
See the search page on OKCupid for a better idea of the
arguments expected.
Parameters
----------
location : string, optional
    Location of profiles returned. Accept ZIP codes, city
    names, city & state combinations, and city & country
    combinations. Default to user location if unable to
    understand the string or if no value is given.
radius : int, optional
    Radius in miles searched, centered on the location.
number : int, optional
    Number of profiles returned. Default to 18, which is the
    same number that OKCupid returns by default.
age_min : int, optional
    Minimum age of profiles returned. Cannot be lower than 18.
age_max : int, optional
    Maximum age of profiles returned. Cannot be higher than 99.
order_by : str, optional
    Order in which profiles are returned.
last_online : str, optional
    How recently online the profiles returned are. Can also be
    an int that represents seconds.
status : str, optional
    Dating status of profiles returned. Default to 'single'
    unless the argument is either 'not single', 'married', or
    'any'.
height_min : int, optional
    Minimum height in inches of profiles returned.
height_max : int, optional
    Maximum height in inches of profiles returned.
looking_for : str, optional
    Describe the gender and orientation of profiles returned.
    If left blank, return some variation of "guys/girls who
    like guys/girls" or "both who like bi girls/guys, depending
    on the user's gender and orientation.
smokes : str or list of str, optional
    Smoking habits of profiles returned.
drinks : str or list of str, optional
    Drinking habits of profiles returned.
drugs : str or list of str, optional
    Drug habits of profiles returned.
education : str or list of str, optional
    Highest level of education attained by profiles returned.
job : str or list of str, optional
    Industry in which the profile users work.
income : str or list of str, optional
    Income range of profiles returned.
religion : str or list of str, optional
    Religion of profiles returned.
monogamy : str or list of str, optional
    Whether the profiles returned are monogamous or non-monogamous.
offspring : str or list of str, optional
    Whether the profiles returned have or want children.
pets : str or list of str, optional
    Dog/cat ownership of profiles returned.
languages : str or list of str, optional
    Languages spoken for profiles returned.
diet : str or list of str, optional
    Dietary restrictions of profiles returned.
sign : str or list of str, optional
    Astrological sign of profiles returned.
ethnicity : str or list of str, optional
    Ethnicity of profiles returned.
join_date : int or str, optional
    Either a string describing the profile join dates ('last
    week', 'last year' etc.) or an int indicating the number
    of maximum seconds from the moment of joining OKCupid.
keywords : str, optional
    Keywords that the profiles returned must contain. Note that
    spaces separate keywords, ie. `keywords="love cats"` will
    return profiles that contain both "love" and "cats" rather
    than the exact string "love cats".
"""
import inspect
import logging

from lxml import html
import simplejson

from . import helpers
from . import magicnumbers
from . import util
from .profile import Profile
from .session import Session
from .xpath import xpb


log = logging.getLogger(__name__)


builder_to_keys = {}
builder_to_decider = {}


def any_decider(function, incoming_keys, accepted_keys):
    return bool(set(incoming_keys).intersection(accepted_keys))


def all_decider(function, incoming_keys, accepted_keys):
    return set(accepted_keys).issubset(incoming_keys)


@util.n_partialable
def register_filter_builder(function, keys=(), decider=all_decider):
    function_arguments = inspect.getargspec(function).args
    if keys:
        assert len(keys) == len(function_arguments)
    else:
        keys = function_arguments
    builder_to_keys[function] = keys
    builder_to_decider[function] = decider

    return function


@register_filter_builder
def looking_for_filter(looking_for):
    return '0,{0}'.format(magicnumbers.seeking[looking_for.lower()])


@register_filter_builder(decider=any_decider)
def age_filter(age_min=18, age_max=99):
    if age_min == None:
        age_min = 18
    return '2,{0},{1}'.format(age_min, age_max)


@register_filter_builder(decider=any_decider)
def attractiveness_filter(attractiveness_min, attractiveness_max):
    if attractiveness_min == None:
        attractiveness_min = 0
    if attractiveness_max == None:
        attractiveness_max = 10000
    return '25,{0},{1}'.format(attractiveness_min, attractiveness_max)


@register_filter_builder(decider=any_decider)
def height_filter(height_min, height_max):
    return magicnumbers.get_height_query(height_min, height_max)


@register_filter_builder
def last_online_filter(last_online):
    return '5,{0}'.format(helpers.format_last_online(last_online))


@register_filter_builder
def status_filter(status):
    status_int = 2  # single, default
    if status.lower() in ('not single', 'married'):
        status_int = 12
    elif status.lower() == 'any':
        status_int = 0
    return '35,{0}'.format(status_int)


@register_filter_builder
def location_filter(radius):
    return '3,{0}'.format(radius)


def build_option_filter(key):
    @register_filter_builder(keys=(key,))
    @util.makelist_decorator
    def option_filter(value):
        return magicnumbers.get_options_query(key, value)



for key in ['smokes', 'drinks', 'drugs', 'education', 'job',
            'income', 'religion', 'monogamy', 'diet', 'sign',
            'ethnicity']:
    build_option_filter(key)

pairs = [('pets', util.makelist_decorator(magicnumbers.get_pet_queries)),
         ('offspring', util.makelist_decorator(magicnumbers.get_kids_query)),
         ('join_date', magicnumbers.get_join_date_query),
         ('languages', magicnumbers.get_language_query)]
for key, function in pairs:

    register_filter_builder(keys=(key,))(function)


class SearchParameterBuilder(object):

    def __init__(self, **kwargs):
        self._options = kwargs

    @property
    def location(self):
        return self._options.get('location', None)

    @property
    def gender(self):
        return self._options.get('gender', 'm')

    @property
    def keywords(self):
        return self._options.get('keywords')

    @property
    def order_by(self):
        return self._options.get('order_by', 'match').upper()

    @property
    def filters(self):
        incoming_keys = tuple(self._options.keys())
        builders = [builder for builder, decider in builder_to_decider.items()
                    if decider(builder, incoming_keys, builder_to_keys[builder])]
        return [builder(*[self._options.get(key)
                          for key in builder_to_keys[builder]])
                for builder in builders]

    def set_options(self, **kwargs):
        self._options.update(kwargs)
        return self._options

    def build(self, session, count=9, low=None):
        search_parameters = {
            'timekey': 1,
            'matchOrderBy': self.order_by,
            'custom_search': '0',
            'fromWhoOnline': '0',
            'mygender': self.gender,
            'update_prefs': '1',
            'sort_type': '0',
            'sa': '1',
            'count': str(count),
            'locid': (str(helpers.get_locid(session, self.location))
                      if self.location else 0),
            'ajax_load': 1,
            'discard_prefs': 1,
            'match_card_class': 'just_appended'
        }
        if low:
            search_parameters['low'] = low
        if self.keywords: search_parameters['keywords'] = self.keywords
        for filter_number, filter_string in enumerate(self.filters, 1):
            search_parameters['filter{0}'.format(filter_number)] = filter_string
        return search_parameters


class MatchCardExtractor(object):

    def __init__(self, div):
        self._div = div

    @property
    def username(self):
        return xpb.div.with_class('username').get_text_(self._div).strip()

    @property
    def age(self):
        return int(xpb.span.with_class('age').get_text_(self._div))

    @property
    def location(self):
        return helpers.replace_chars(
            xpb.span.with_class('location').get_text_(self._div)
        )

    _match_percentage_xpb = xpb.div.with_classes('percentage_wrapper', 'match').\
                            span.with_classes('percentage')

    @property
    def match_percentage(self):
        try:
            return int(self._match_percentage_xpb.get_text_(self._div).strip('%'))
        except:
            return 0

    _enemy_percentage_xpb = xpb.div.with_classes('percentage_wrapper', 'enemy').\
                            span.with_classes('percentage')

    @property
    def enemy_percentage(self):
        try:
            return int(self._enemy_percentage_xpb.get_text_(self._div).strip('%'))
        except ValueError:
            return 0

    @property
    def contacted(self):
        return bool(xpb.div.with_class('fancydate').apply_(self._div))

    @property
    def as_dict(self):
        return {
            'username': self.username,
            'age': self.age,
            'location': self.location,
            'match_percentage': self.match_percentage,
            'enemy_percentage': self.enemy_percentage,
            'id': self.id,
            'rating': self.rating,
            'contacted': self.contacted
        }


class SearchManager(object):

    def __init__(self, session=None, low=1, **kwargs):
        self._session = session or Session.login()
        self._parameter_builder = SearchParameterBuilder(**kwargs)
        self.set_options = self._parameter_builder.set_options
        self._low = low

    def get_search_html(self, count=9):
        search_parameters = self._parameter_builder.build(self._session,
                                                          count, self._low)
        log.info(simplejson.dumps({'search_parameters': search_parameters}))
        response = self._session.get('https://www.okcupid.com/match',
                                     params=search_parameters)
        try:
            search_html = response.json()['html']
        except:
            log.warning(simplejson.dumps(
                {'failure': response.content.decode('utf8')}
            ))
            raise
        return search_html

    def get_profiles(self, count=9):
        search_html = self.get_search_html(count)
        if not search_html.strip(): return []
        tree = html.fromstring(search_html)
        match_card_elems = xpb.div.with_classes('match_card').apply_(tree)

        profiles = []
        for div in match_card_elems:
            match_card_extractor = MatchCardExtractor(div)
            profiles.append(Profile(self._session, match_card_extractor.username))
        self._low += len(profiles)
        return profiles

    def __iter__(self):
        profiles = self.get_profiles()
        while profiles:
            for profile in profiles:
                yield profile
            profiles = self.get_profiles()


def search(session=None, count=9, **kwargs):
    return SearchManager(session, **kwargs).get_profiles(count)



from okcupyd.db import adapters


class MailboxSyncer(object):

    def __init__(self, mailbox):
        self._mailbox = mailbox

    def sync(self):
        threads = self._mailbox.refresh()
        for thread in threads:
            adapters.ThreadAdapter(thread).get_thread()



import operator

from okcupyd import util


@util.n_partialable
def crazy(a, self):
    return a + self.starter


class HasPartialableMethods(object):

    def __init__(self, starter):
        self.starter = starter
        self.on_instance = crazy

    @util.n_partialable
    def partialable(self, x, y=0):
        return self.starter * x * y

    on_class = crazy(1)


def test_partialable_method_behavior():
    assert HasPartialableMethods(2).partialable(y=2)(1) == 4
    hpm = HasPartialableMethods(2)
    assert hpm.on_instance(2, hpm) == 4
    assert hpm.on_class() == 3


def test_partialable_with_kwargs_taking_function():
    @util.n_partialable
    def kwarg_taking_function(arg, **kwargs):
        kwargs['k'] = arg
        return kwargs

    assert kwarg_taking_function(a=14)(2) == {'k': 2, 'a': 14}


def test_get_cached_properties():
    class PropClass(object):
        def __init__(self):
            self.count = 0

        @util.cached_property
        def count_prop(self):
            self.count += 1
            return self.count

        @util.cached_property
        def other_prop(self):
            return 4

    instance = PropClass()
    assert ['count_prop', 'other_prop'] == \
        list(map(operator.itemgetter(0),
                 util.cached_property.get_cached_properties(instance)))

    assert instance.count_prop == 1

    util.cached_property.bust_caches(instance)

    assert instance.count_prop == 2


def test_overwrite_kwarg():

    @util.n_partialable
    def test(k, a=0, b=0):
        return k + a + b

    assert test(a=4)(a=1)(1) == 2



# -*- coding: utf-8 -*-
from collections import namedtuple
from datetime import datetime, date
from json import loads
from lxml import html, etree
from re import search
import logging

import simplejson

from .xpath import xpb
from . import util


CHAR_REPLACE = {
    u"′": "'",
    u'″': '"',
    u'“': '"',
    u'”': '"',
    u"’": "'",
    u"—": "-",
    u"–": "-",
    u"…": "...",
    u"🌲": " ",
    u'\u2019': "'",
}


log = logging.getLogger(__name__)


MessageInfo = namedtuple('MessageInfo', ('thread_id', 'message_id'))


class Messager(object):

    def __init__(self, session):
        self._session = session

    def _get_authcode(self, username):
        response = self._session.get(
            'https://www.okcupid.com/profile/{0}'.format(username)
        ).content.decode('utf8')
        return get_authcode(response)

    def message_request_parameters(self, username, message, thread_id, authcode):
        return {
            'ajax': 1,
            'sendmsg': 1,
            'r1': username,
            'body': message,
            'threadid': thread_id,
            'authcode': authcode,
            'reply': 1 if thread_id else 0,
            'from_profile': 1
        }

    def send(self, username, message, authcode=None, thread_id=None):
        authcode = authcode or self._get_authcode(username)
        params = self.message_request_parameters(username, message,
                                                 thread_id or 0, authcode)
        response = self._session.get('https://www.okcupid.com/mailbox', params=params)
        response_dict = response.json()
        log.info(simplejson.dumps({'message_send_response': response_dict}))
        return MessageInfo(response_dict.get('threadid'), response_dict['msgid'])


def get_additional_info(tree):
    age = int(xpb.span(id='ajax_age').get_text_(tree).strip())
    orientation = tree.xpath("//*[@id = 'ajax_orientation']/text()")[0].strip()
    status = tree.xpath("//*[@id = 'ajax_status']/text()")[0].strip()
    gender_result = tree.xpath("//*[@id = 'ajax_gender']/text()")[0].strip()
    location = tree.xpath("//*[@id = 'ajax_location']/text()")[0].strip()
    if gender_result == "M":
        gender = 'Male'
    elif gender_result == "F":
        gender = 'Female'
    return age, gender, orientation, status, location


utf8_parser = etree.XMLParser(encoding='utf-8')


@util.n_partialable
def get_js_variable(html_response, variable_name):
    script_elements = xpb.script.apply_(html_response)
    html_response = '\n'.join(script_element.text_content()
                              for script_element in script_elements)
    return search('var {0} = "(.*?)";'.format(variable_name), html_response).group(1)


get_authcode = get_js_variable(variable_name='AUTHCODE')
get_username = get_js_variable(variable_name='SCREENNAME')
get_id = get_js_variable(variable_name='CURRENTUSERID')
get_current_user_id = get_id
get_my_username = get_username


def parse_date_updated(date_updated_text):
    if '/' in date_updated_text:
        return datetime.strptime(date_updated_text, '%m/%d/%y').date()

    if date_updated_text[-2] == ' ':
        month, day = date_updated_text.split()
        date_updated_text = '{0} 0{1}'.format(month, day)

    try:
        return datetime.strptime(date_updated_text, '%b %d').replace(year=datetime.now().year).date()
    except:
        return date.today()


def get_locid(session, location):
    """
    Make a request to locquery resource to translate a string location
    search into an int locid.
    Returns
    ----------
    int
        An int that OKCupid maps to a particular geographical location.
    """
    locid = 0
    query_parameters = {
        'func': 'query',
        'query': location,
    }
    loc_query = session.get('http://www.okcupid.com/locquery', params=query_parameters)
    p = html.fromstring(loc_query.content.decode('utf8'))
    js = loads(p.text)
    if 'results' in js and len(js['results']):
        locid = js['results'][0]['locid']
    return locid


def format_last_online(last_online):
    """
    Return the upper limit in seconds that a profile may have been
    online. If last_online is an int, return that int. Otherwise if
    last_online is a str, convert the string into an int.
    Returns
    ----------
    int
    """
    if isinstance(last_online, str):
        if last_online.lower() in ('day', 'today'):
            last_online_int = 86400  # 3600 * 24
        elif last_online.lower() == 'week':
            last_online_int = 604800  # 3600 * 24 * 7
        elif last_online.lower() == 'month':
            last_online_int = 2678400  # 3600 * 24 * 31
        elif last_online.lower() == 'year':
            last_online_int = 31536000  # 3600 * 365
        elif last_online.lower() == 'decade':
            last_online_int = 315360000  # 3600 * 365 * 10
        else: # Defaults any other strings to last hour
            last_online_int = 3600
    else:
        last_online_int = last_online
    return last_online_int


def update_looking_for(profile_tree, looking_for):
    """
    Update looking_for attribute of a Profile.
    """
    div = profile_tree.xpath("//div[@id = 'what_i_want']")[0]
    looking_for['gentation'] = div.xpath(".//li[@id = 'ajax_gentation']/text()")[0].strip()
    looking_for['ages'] = replace_chars(div.xpath(".//li[@id = 'ajax_ages']/text()")[0].strip())
    looking_for['near'] = div.xpath(".//li[@id = 'ajax_near']/text()")[0].strip()
    looking_for['single'] = div.xpath(".//li[@id = 'ajax_single']/text()")[0].strip()
    try:
        looking_for['seeking'] = div.xpath(".//li[@id = 'ajax_lookingfor']/text()")[0].strip()
    except:
        pass


def update_details(profile_tree, details):
    """
    Update details attribute of a Profile.
    """
    div = profile_tree.xpath("//div[@id = 'profile_details']")[0]
    for dl in div.iter('dl'):
        title = dl.find('dt').text
        item = dl.find('dd')
        if title == 'Last Online' and item.find('span') is not None:
            details[title.lower()] = item.find('span').text.strip()
        elif title.lower() in details and len(item.text):
            details[title.lower()] = item.text.strip()
        else:
            continue
        details[title.lower()] = replace_chars(details[title.lower()])


def get_looking_for(gender, orientation):
    """
    Return a string containing the default gender/orientation of a
    search if no value has been provided to the looking_for kwarg.
    Returns
    ----------
    str
    """
    looking_for = ''
    if gender.lower() in ('male', 'm',):
        if orientation == 'Straight':
            looking_for = 'girls who like guys'
        elif orientation == 'Gay':
            looking_for = 'guys who like guys'
        elif orientation == 'Bisexual':
            looking_for = 'both who like bi guys'
    elif gender.loiwer() in ('female', 'f'):
        if orientation == 'Straight':
            looking_for = 'guys who like girls'
        elif orientation == 'Gay':
            looking_for = 'girls who like girls'
        elif orientation == 'Bisexual':
            looking_for = 'both who like bi girls'
    return looking_for


def replace_chars(astring):
    """
    Replace certain unicode characters to avoid errors when trying
    to read various strings.
    Returns
    ----------
    str
    """
    for k, v in CHAR_REPLACE.items():
        astring = astring.replace(k, v)
    return astring


def add_newlines(tree):
    """
    Add a newline character to the end of each <br> element.
    """
    for br in tree.xpath("*//br"):
        br.tail = "\n" + br.tail if br.tail else "\n"



import argparse
import os
import pkg_resources
import sys

from IPython import start_ipython

from . import util

def parse_args_and_run():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--version', action='store_true',
                        help='Display the version number of okcupyd')
    util.add_command_line_options(parser.add_argument)
    args = parser.parse_args()
    if args.version:
        print(pkg_resources.get_distribution('okcupyd').version)
        sys.exit()
    util.handle_command_line_options(args)
    util.get_credentials()
    sys.exit(start_ipython(['-i', os.path.join(os.path.dirname(__file__), 'start.py')]))


from .attractiveness_finder import AttractivenessFinder
from .photo import PhotoUploader
from .search import search
from .session import Session
from .statistics import Statistics
from .user import User
from .util import save_file


__all__ = ('search', 'User', 'AttractivenessFinder', 'Statistics',
           'save_file', 'parse_args_and_run', 'PhotoUploader', 'Session')



import logging

from lxml import html
import simplejson

from . import util
from . import helpers
from .session import Session
from .xpath import xpb


log = logging.getLogger(__name__)


@util.n_partialable
class PhotoUploader(object):

    _uri = 'https://www.okcupid.com/ajaxuploader'

    def __init__(self, filename, session=None, user_id=None, authcode=None):
        self._session = session or Session.login()
        self._filename = filename
        if not (user_id and authcode):
            photo_tree = html.fromstring(self._session.okc_get(
                'profile/{0}/photos#upload'.format(self._session.log_in_name)
            ).content)
        self._authcode = authcode or helpers.get_authcode(photo_tree)
        self._user_id = user_id or helpers.get_current_user_id(photo_tree)

    @property
    def pic_id(self):
        return self._response_dict['id']

    @property
    def height(self):
        return self._response_dict['height']

    @property
    def width(self):
        return self._response_dict['width']

    def upload(self):
        with open(self._filename, 'rb') as file_object:
            files = {'file': (self._filename, file_object,
                              'image/jpeg', {'Expires': '0'})}
            response = self._session.post(self._uri, files=files)
            response_script_text = xpb.script.get_text_(
                html.fromstring(response.content)
            )
            self._response_dict = self._get_response_json(response_script_text)
            return self._response_dict

    def _get_response_json(self, response_text):
        start = response_text.find('res =') + 5
        end = response_text.find('};') + 1
        response_text = response_text[start:end]
        log.info(simplejson.dumps({'photo_upload_response': response_text}))
        return simplejson.loads(response_text)

    @property
    def _confirm_parameters(self):
        return {
            'userid': self._user_id,
            'albumid': 0,
            'authcode': self._authcode,
            'okc_api': 1,
            'picid': self.pic_id,
            'picture.add_ajax': 1,
            'use_new_upload': 1,
            'caption': 1,
            'height': self.height,
            'width': self.width,
        }

    def confirm(self):
        return self._session.okc_get('photoupload',
                                     params=self._confirm_parameters)

    def upload_and_confirm(self):
        self.upload()
        self.confirm()
        return self._response_dict



from lxml import html

from . import helpers
from . import util
from .messaging import ThreadFetcher
from .photo import PhotoUploader
from .profile import Profile
from .search import SearchManager
from .session import Session
from .xpath import xpb


class User(object):

    @classmethod
    def with_credentials(cls, username, password):
        return cls(Session.login(username, password))

    def __init__(self, session=None):
        self._session = session or Session.login()
        assert self._session.log_in_name is not None, (
            "The session provided to the user constructor must be logged in."
        )

        self.profile = Profile(self._session, self._session.log_in_name)
        self.authcode = self.profile.authcode

        self._message_sender = helpers.Messager(self._session)
        self.inbox = util.Fetchable(ThreadFetcher(self._session, 1))
        self.outbox = util.Fetchable(ThreadFetcher(self._session, 2))
        self.drafts = util.Fetchable(ThreadFetcher(self._session, 4))

    def get_profile(self, username):
        return self._session.get_profile(username)

    _visitors_xpb = xpb.div.with_class('user_info').\
                   div.with_class('profile_info').div.with_class('username').\
                   a.with_class('name')

    _visitors_xpb = xpb.div.with_class('username').\
                   a.with_class('name')

    @util.cached_property
    def visitors(self):
        visitors_response = self._session.okc_get('visitors').content.decode('utf8')
        visitors_tree = html.fromstring(visitors_response)
        return [Profile(self._session, username)
                for username in self._visitors_xpb.text_.apply_(visitors_tree)]

    def message(self, username, message_text):
        # Try to reply to an existing thread.
        if not isinstance(username, str):
            username = username.username
        for thread in sorted(set(self.inbox.items + self.outbox.items),
                             key=lambda t: t.date, reverse=True):
            if thread.correspondent.lower() == username.lower():
                thread.reply(message_text)
                return

        return self._message_sender.send(username, message_text)

    def search(self, **kwargs):
        count = kwargs.pop('count', 9)
        return self.search_manager(**kwargs).get_profiles(count=count)

    def search_manager(self, **kwargs):
        kwargs.setdefault('gender', self.profile.gender[0])
        looking_for = helpers.get_looking_for(self.profile.gender,
                                              self.profile.orientation)
        kwargs.setdefault('looking_for', looking_for)
        kwargs.setdefault('location', self.profile.location)
        kwargs.setdefault('radius', 25)
        return SearchManager(session=self._session, **kwargs)

    def quickmatch(self):
        """Return a Profile obtained by visiting the quickmatch page."""
        response = self._session.okc_get('quickmatch', params={'okc_api': 1})
        return Profile(self._session, response.json()['sn'])

    def upload_photo(self, filename):
        return PhotoUploader(filename, self._session,
                             user_id=self.profile._current_user_id)

    def __repr__(self):
        return 'User("{0}")'.format(self.profile.username)
